# 1 "/Users/charger/src/kudu/python/116/python/kudu/config.pxi.in"
# 1 "<built-in>" 1
# 1 "<built-in>" 3
# 382 "<built-in>" 3
# 1 "<command line>" 1
# 1 "<built-in>" 2
# 1 "/Users/charger/src/kudu/src/kudu/util/int128.h" 1
# 2 "<built-in>" 2
# 1 "/Users/charger/src/kudu/python/116/python/kudu/config.pxi.in" 2
# 23 "/Users/charger/src/kudu/python/116/python/kudu/config.pxi.in"
DEF PYKUDU_INT128_SUPPORTED = 1
